@@ -17,9 +17,11 @@
 #include "access/htup_details.h"
 #include "access/xact.h"
 #include "catalog/indexing.h"
+#include "catalog/pg_type.h"
 #include "commands/sequence.h"
 #include "distributed/colocation_utils.h"
 #include "distributed/listutils.h"
+#include "distributed/master_metadata_utility.h"
 #include "distributed/master_protocol.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/metadata_sync.h"
@@ -29,6 +31,7 @@
 #include "distributed/shardinterval_utils.h"
 #include "distributed/worker_protocol.h"
 #include "distributed/worker_transaction.h"
+#include "storage/lmgr.h"
 #include "utils/builtins.h"
 #include "utils/fmgroids.h"
 #include "utils/lsyscache.h"
@@ -51,6 +54,7 @@ static void DeleteColocationGroup(uint32 colocationId);
 
 /* exports for SQL callable functions */
 PG_FUNCTION_INFO_V1(mark_tables_colocated);
+PG_FUNCTION_INFO_V1(get_colocated_shard_array);
 
 
 /*
@@ -90,6 +94,43 @@ mark_tables_colocated(PG_FUNCTION_ARGS)
 }
 
 
+/*
+ * get_colocated_shards_array returns array of shards ids which are co-located with given
+ * shard.
+ */
+Datum
+get_colocated_shard_array(PG_FUNCTION_ARGS)
+{
+	uint32 shardId = PG_GETARG_UINT32(0);
+	ShardInterval *shardInterval = LoadShardInterval(shardId);
+
+	ArrayType *colocatedShardsArrayType = NULL;
+	List *colocatedShardList = ColocatedShardIntervalList(shardInterval);
+	ListCell *colocatedShardCell = NULL;
+	int colocatedShardCount = list_length(colocatedShardList);
+	Datum *colocatedShardsDatumArray = palloc0(colocatedShardCount * sizeof(Datum));
+	Oid arrayTypeId = OIDOID;
+	int colocatedShardIndex = 0;
+
+	foreach(colocatedShardCell, colocatedShardList)
+	{
+		ShardInterval *colocatedShardInterval = (ShardInterval *) lfirst(
+			colocatedShardCell);
+		uint64 colocatedShardId = colocatedShardInterval->shardId;
+
+		Datum colocatedShardDatum = Int64GetDatum(colocatedShardId);
+
+		colocatedShardsDatumArray[colocatedShardIndex] = colocatedShardDatum;
+		colocatedShardIndex++;
+	}
+
+	colocatedShardsArrayType = DatumArrayToArrayType(colocatedShardsDatumArray,
+													 colocatedShardCount, arrayTypeId);
+
+	PG_RETURN_ARRAYTYPE_P(colocatedShardsArrayType);
+}
+
+
 /*
  * MarkTablesColocated puts both tables to same colocation group. If the
  * source table is in INVALID_COLOCATION_ID group, then it creates a new
@@ -126,7 +167,7 @@ MarkTablesColocated(Oid sourceRelationId, Oid targetRelationId)
 		uint32 shardCount = ShardIntervalCount(sourceRelationId);
 		uint32 shardReplicationFactor = TableShardReplicationFactor(sourceRelationId);
 
-		Var *sourceDistributionColumn = PartitionKey(sourceRelationId);
+		Var *sourceDistributionColumn = DistPartitionKey(sourceRelationId);
 		Oid sourceDistributionColumnType = InvalidOid;
 
 		/* reference tables has NULL distribution column */
@@ -477,8 +518,7 @@ CreateColocationGroup(int shardCount, int replicationFactor, Oid distributionCol
 	tupleDescriptor = RelationGetDescr(pgDistColocation);
 	heapTuple = heap_form_tuple(tupleDescriptor, values, isNulls);
 
-	simple_heap_insert(pgDistColocation, heapTuple);
-	CatalogUpdateIndexes(pgDistColocation, heapTuple);
+	CatalogTupleInsert(pgDistColocation, heapTuple);
 
 	/* increment the counter so that next command can see the row */
 	CommandCounterIncrement();
@@ -567,7 +607,7 @@ CheckDistributionColumnType(Oid sourceRelationId, Oid targetRelationId)
 	Oid targetDistributionColumnType = InvalidOid;
 
 	/* reference tables have NULL distribution column */
-	sourceDistributionColumn = PartitionKey(sourceRelationId);
+	sourceDistributionColumn = DistPartitionKey(sourceRelationId);
 	if (sourceDistributionColumn == NULL)
 	{
 		sourceDistributionColumnType = InvalidOid;
@@ -578,7 +618,7 @@ CheckDistributionColumnType(Oid sourceRelationId, Oid targetRelationId)
 	}
 
 	/* reference tables have NULL distribution column */
-	targetDistributionColumn = PartitionKey(targetRelationId);
+	targetDistributionColumn = DistPartitionKey(targetRelationId);
 	if (targetDistributionColumn == NULL)
 	{
 		targetDistributionColumnType = InvalidOid;
@@ -648,9 +688,10 @@ UpdateRelationColocationGroup(Oid distributedRelationId, uint32 colocationId)
 	replace[Anum_pg_dist_partition_colocationid - 1] = true;
 
 	heapTuple = heap_modify_tuple(heapTuple, tupleDescriptor, values, isNull, replace);
-	simple_heap_update(pgDistPartition, &heapTuple->t_self, heapTuple);
 
-	CatalogUpdateIndexes(pgDistPartition, heapTuple);
+
+	CatalogTupleUpdate(pgDistPartition, &heapTuple->t_self, heapTuple);
+
 	CitusInvalidateRelcacheByRelid(distributedRelationId);
 
 	CommandCounterIncrement();
@@ -889,6 +930,9 @@ ColocatedShardIntervalList(ShardInterval *shardInterval)
 /*
  * ColocatedTableId returns an arbitrary table which belongs to given colocation
  * group. If there is not such a colocation group, it returns invalid oid.
+ *
+ * This function also takes an AccessShareLock on the co-colocated table to
+ * guarantee that the table isn't dropped for the remainder of the transaction.
  */
 Oid
 ColocatedTableId(Oid colocationId)
@@ -903,10 +947,19 @@ ColocatedTableId(Oid colocationId)
 	ScanKeyData scanKey[1];
 	int scanKeyCount = 1;
 
+	/*
+	 * We may have a distributed table whose colocation id is INVALID_COLOCATION_ID.
+	 * In this case, we do not want to send that table's id as colocated table id.
+	 */
+	if (colocationId == INVALID_COLOCATION_ID)
+	{
+		return colocatedTableId;
+	}
+
 	ScanKeyInit(&scanKey[0], Anum_pg_dist_partition_colocationid,
 				BTEqualStrategyNumber, F_INT4EQ, ObjectIdGetDatum(colocationId));
 
-	/* prevent DELETE statements */
+	/* do not allow any tables to be dropped while we read from pg_dist_partition */
 	pgDistPartition = heap_open(DistPartitionRelationId(), ShareLock);
 	tupleDescriptor = RelationGetDescr(pgDistPartition);
 	scanDescriptor = systable_beginscan(pgDistPartition,
@@ -918,6 +971,9 @@ ColocatedTableId(Oid colocationId)
 	{
 		colocatedTableId = heap_getattr(heapTuple, Anum_pg_dist_partition_logicalrelid,
 										tupleDescriptor, &isNull);
+
+		/* make sure the table isn't dropped for the remainder of the transaction */
+		LockRelationOid(colocatedTableId, AccessShareLock);
 	}
 
 	systable_endscan(scanDescriptor);
