@@ -34,12 +34,16 @@
 #include "distributed/pg_dist_node.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/pg_dist_shard.h"
-#include "distributed/pg_dist_shard_placement.h"
+#include "distributed/pg_dist_placement.h"
+#include "distributed/shared_library_init.h"
 #include "distributed/shardinterval_utils.h"
 #include "distributed/worker_manager.h"
 #include "distributed/worker_protocol.h"
 #include "executor/executor.h"
+#include "nodes/makefuncs.h"
 #include "parser/parse_func.h"
+#include "parser/parse_type.h"
+#include "storage/lmgr.h"
 #include "utils/builtins.h"
 #include "utils/catcache.h"
 #include "utils/datum.h"
@@ -54,6 +58,9 @@
 #include "utils/typcache.h"
 
 
+/* user configuration */
+int ReadFromSecondaries = USE_SECONDARY_NODES_NEVER;
+
 /*
  * ShardCacheEntry represents an entry in the shardId -> ShardInterval cache.
  * To avoid duplicating data and invalidation logic between this cache and the
@@ -81,27 +88,41 @@ typedef struct ShardCacheEntry
 } ShardCacheEntry;
 
 
-/* state which should be cleared upon DROP EXTENSION */
-static bool extensionLoaded = false;
-static Oid distShardRelationId = InvalidOid;
-static Oid distShardPlacementRelationId = InvalidOid;
-static Oid distNodeRelationId = InvalidOid;
-static Oid distLocalGroupRelationId = InvalidOid;
-static Oid distColocationRelationId = InvalidOid;
-static Oid distColocationConfigurationIndexId = InvalidOid;
-static Oid distColocationColocationidIndexId = InvalidOid;
-static Oid distPartitionRelationId = InvalidOid;
-static Oid distPartitionLogicalRelidIndexId = InvalidOid;
-static Oid distPartitionColocationidIndexId = InvalidOid;
-static Oid distShardLogicalRelidIndexId = InvalidOid;
-static Oid distShardShardidIndexId = InvalidOid;
-static Oid distShardPlacementShardidIndexId = InvalidOid;
-static Oid distShardPlacementPlacementidIndexId = InvalidOid;
-static Oid distShardPlacementNodeidIndexId = InvalidOid;
-static Oid distTransactionRelationId = InvalidOid;
-static Oid distTransactionGroupIndexId = InvalidOid;
-static Oid extraDataContainerFuncId = InvalidOid;
-static Oid workerHashFunctionId = InvalidOid;
+/*
+ * State which should be cleared upon DROP EXTENSION.  When the configuration
+ * changes, e.g. because the extension is dropped, these summarily get set to
+ * 0.
+ */
+typedef struct MetadataCacheData
+{
+	bool extensionLoaded;
+	Oid distShardRelationId;
+	Oid distPlacementRelationId;
+	Oid distNodeRelationId;
+	Oid distLocalGroupRelationId;
+	Oid distColocationRelationId;
+	Oid distColocationConfigurationIndexId;
+	Oid distColocationColocationidIndexId;
+	Oid distPartitionRelationId;
+	Oid distPartitionLogicalRelidIndexId;
+	Oid distPartitionColocationidIndexId;
+	Oid distShardLogicalRelidIndexId;
+	Oid distShardShardidIndexId;
+	Oid distPlacementShardidIndexId;
+	Oid distPlacementPlacementidIndexId;
+	Oid distPlacementGroupidIndexId;
+	Oid distTransactionRelationId;
+	Oid distTransactionGroupIndexId;
+	Oid extraDataContainerFuncId;
+	Oid workerHashFunctionId;
+	Oid extensionOwner;
+	Oid primaryNodeRoleId;
+	Oid secondaryNodeRoleId;
+	Oid unavailableNodeRoleId;
+} MetadataCacheData;
+
+
+static MetadataCacheData MetadataCache;
 
 /* Citus extension version variables */
 bool EnableVersionChecks = true; /* version checks are enabled */
@@ -118,8 +139,6 @@ static HTAB *DistShardCacheHash = NULL;
 static HTAB *WorkerNodeHash = NULL;
 static bool workerNodeHashValid = false;
 
-static bool invalidationRegistered = false;
-
 /* default value is -1, for coordinator it's 0 and for worker nodes > 0 */
 static int LocalGroupId = -1;
 
@@ -148,8 +167,11 @@ static char * InstalledExtensionVersion(void);
 static bool HasOverlappingShardInterval(ShardInterval **shardIntervalArray,
 										int shardIntervalArrayLength,
 										FmgrInfo *shardIntervalSortCompareFunction);
+static void InitializeCaches(void);
 static void InitializeDistTableCache(void);
 static void InitializeWorkerNodeCache(void);
+static void RegisterWorkerNodeCacheCallbacks(void);
+static void RegisterLocalGroupIdCacheCallbacks(void);
 static uint32 WorkerNodeHashCode(const void *key, Size keySize);
 static void ResetDistTableCacheEntry(DistTableCacheEntry *cacheEntry);
 static void InvalidateDistRelationCacheCallback(Datum argument, Oid relationId);
@@ -165,6 +187,9 @@ static ShardInterval * TupleToShardInterval(HeapTuple heapTuple,
 											TupleDesc tupleDescriptor, Oid intervalTypeId,
 											int32 intervalTypeMod);
 static void CachedRelationLookup(const char *relationName, Oid *cachedOid);
+static ShardPlacement * ResolveGroupShardPlacement(
+	GroupShardPlacement *groupShardPlacement, ShardCacheEntry *shardEntry);
+static WorkerNode * LookupNodeForGroup(uint32 groupid);
 
 
 /* exports for SQL callable functions */
@@ -175,6 +200,27 @@ PG_FUNCTION_INFO_V1(master_dist_node_cache_invalidate);
 PG_FUNCTION_INFO_V1(master_dist_local_group_cache_invalidate);
 
 
+/*
+ * EnsureModificationsCanRun checks if the current node is in recovery mode or
+ * citus.use_secondary_nodes is 'alwaus'. If either is true the function errors out.
+ */
+void
+EnsureModificationsCanRun(void)
+{
+	if (RecoveryInProgress())
+	{
+		ereport(ERROR, (errmsg("writing to worker nodes is not currently allowed"),
+						errdetail("the database is in recovery mode")));
+	}
+
+	if (ReadFromSecondaries == USE_SECONDARY_NODES_ALWAYS)
+	{
+		ereport(ERROR, (errmsg("writing to worker nodes is not currently allowed"),
+						errdetail("citus.use_secondary_nodes is set to 'always'")));
+	}
+}
+
+
 /*
  * IsDistributedTable returns whether relationId is a distributed relation or
  * not.
@@ -301,18 +347,18 @@ LoadShardInterval(uint64 shardId)
 
 
 /*
- * LoadShardPlacement returns the, cached, metadata about a shard placement.
+ * LoadGroupShardPlacement returns the cached shard placement metadata
  *
- * The return value is a copy of the cached ShardPlacement struct and may
+ * The return value is a copy of the cached GroupShardPlacement struct and may
  * therefore be modified and/or freed.
  */
-ShardPlacement *
-LoadShardPlacement(uint64 shardId, uint64 placementId)
+GroupShardPlacement *
+LoadGroupShardPlacement(uint64 shardId, uint64 placementId)
 {
 	ShardCacheEntry *shardEntry = NULL;
 	DistTableCacheEntry *tableEntry = NULL;
 
-	ShardPlacement *placementArray = NULL;
+	GroupShardPlacement *placementArray = NULL;
 	int numberOfPlacements = 0;
 
 	int i = 0;
@@ -330,8 +376,9 @@ LoadShardPlacement(uint64 shardId, uint64 placementId)
 	{
 		if (placementArray[i].placementId == placementId)
 		{
-			ShardPlacement *shardPlacement = CitusMakeNode(ShardPlacement);
-			CopyShardPlacement(&placementArray[i], shardPlacement);
+			GroupShardPlacement *shardPlacement = CitusMakeNode(GroupShardPlacement);
+
+			memcpy(shardPlacement, &placementArray[i], sizeof(GroupShardPlacement));
 
 			return shardPlacement;
 		}
@@ -342,6 +389,169 @@ LoadShardPlacement(uint64 shardId, uint64 placementId)
 }
 
 
+/*
+ * LoadShardPlacement returns a shard placement for the primary node.
+ */
+ShardPlacement *
+LoadShardPlacement(uint64 shardId, uint64 placementId)
+{
+	ShardCacheEntry *shardEntry = NULL;
+	GroupShardPlacement *groupPlacement = NULL;
+	ShardPlacement *nodePlacement = NULL;
+
+	shardEntry = LookupShardCacheEntry(shardId);
+	groupPlacement = LoadGroupShardPlacement(shardId, placementId);
+	nodePlacement = ResolveGroupShardPlacement(groupPlacement, shardEntry);
+
+	return nodePlacement;
+}
+
+
+/*
+ * FindShardPlacementOnGroup returns the shard placement for the given shard
+ * on the given group, or returns NULL of no placement for the shard exists
+ * on the group.
+ */
+ShardPlacement *
+FindShardPlacementOnGroup(uint32 groupId, uint64 shardId)
+{
+	ShardCacheEntry *shardEntry = NULL;
+	DistTableCacheEntry *tableEntry = NULL;
+	GroupShardPlacement *placementArray = NULL;
+	int numberOfPlacements = 0;
+	ShardPlacement *placementOnNode = NULL;
+	int placementIndex = 0;
+
+	shardEntry = LookupShardCacheEntry(shardId);
+	tableEntry = shardEntry->tableEntry;
+	placementArray = tableEntry->arrayOfPlacementArrays[shardEntry->shardIndex];
+	numberOfPlacements = tableEntry->arrayOfPlacementArrayLengths[shardEntry->shardIndex];
+
+	for (placementIndex = 0; placementIndex < numberOfPlacements; placementIndex++)
+	{
+		GroupShardPlacement *placement = &placementArray[placementIndex];
+
+		if (placement->groupId == groupId)
+		{
+			placementOnNode = ResolveGroupShardPlacement(placement, shardEntry);
+			break;
+		}
+	}
+
+	return placementOnNode;
+}
+
+
+/*
+ * ResolveGroupShardPlacement takes a GroupShardPlacement and adds additional data to it,
+ * such as the node we should consider it to be on.
+ */
+static ShardPlacement *
+ResolveGroupShardPlacement(GroupShardPlacement *groupShardPlacement,
+						   ShardCacheEntry *shardEntry)
+{
+	DistTableCacheEntry *tableEntry = shardEntry->tableEntry;
+	int shardIndex = shardEntry->shardIndex;
+	ShardInterval *shardInterval = tableEntry->sortedShardIntervalArray[shardIndex];
+
+	ShardPlacement *shardPlacement = CitusMakeNode(ShardPlacement);
+	uint32 groupId = groupShardPlacement->groupId;
+	WorkerNode *workerNode = LookupNodeForGroup(groupId);
+
+	/* copy everything into shardPlacement but preserve the header */
+	memcpy((((CitusNode *) shardPlacement) + 1),
+		   (((CitusNode *) groupShardPlacement) + 1),
+		   sizeof(GroupShardPlacement) - sizeof(CitusNode));
+
+	shardPlacement->nodeName = pstrdup(workerNode->workerName);
+	shardPlacement->nodePort = workerNode->workerPort;
+
+	/* fill in remaining fields */
+	Assert(tableEntry->partitionMethod != 0);
+	shardPlacement->partitionMethod = tableEntry->partitionMethod;
+	shardPlacement->colocationGroupId = tableEntry->colocationId;
+	if (tableEntry->partitionMethod == DISTRIBUTE_BY_HASH)
+	{
+		Assert(shardInterval->minValueExists);
+		Assert(shardInterval->valueTypeId == INT4OID);
+
+		/*
+		 * Use the lower boundary of the interval's range to identify
+		 * it for colocation purposes. That remains meaningful even if
+		 * a concurrent session splits a shard.
+		 */
+		shardPlacement->representativeValue = DatumGetInt32(shardInterval->minValue);
+	}
+	else
+	{
+		shardPlacement->representativeValue = 0;
+	}
+
+	return shardPlacement;
+}
+
+
+/*
+ * LookupNodeForGroup searches the WorkerNodeHash for a worker which is a member of the
+ * given group and also readable (a primary if we're reading from primaries, a secondary
+ * if we're reading from secondaries). If such a node does not exist it emits an
+ * appropriate error message.
+ */
+static WorkerNode *
+LookupNodeForGroup(uint32 groupId)
+{
+	WorkerNode *workerNode = NULL;
+	HASH_SEQ_STATUS status;
+	HTAB *workerNodeHash = GetWorkerNodeHash();
+	bool foundAnyNodes = false;
+
+	hash_seq_init(&status, workerNodeHash);
+
+	while ((workerNode = hash_seq_search(&status)) != NULL)
+	{
+		uint32 workerNodeGroupId = workerNode->groupId;
+		if (workerNodeGroupId != groupId)
+		{
+			continue;
+		}
+
+		foundAnyNodes = true;
+
+		if (WorkerNodeIsReadable(workerNode))
+		{
+			hash_seq_term(&status);
+			return workerNode;
+		}
+	}
+
+	if (!foundAnyNodes)
+	{
+		ereport(ERROR, (errmsg("there is a shard placement in node group %u but "
+							   "there are no nodes in that group", groupId)));
+	}
+
+	switch (ReadFromSecondaries)
+	{
+		case USE_SECONDARY_NODES_NEVER:
+		{
+			ereport(ERROR, (errmsg("node group %u does not have a primary node",
+								   groupId)));
+		}
+
+		case USE_SECONDARY_NODES_ALWAYS:
+		{
+			ereport(ERROR, (errmsg("node group %u does not have a secondary node",
+								   groupId)));
+		}
+
+		default:
+		{
+			ereport(FATAL, (errmsg("unrecognized value for use_secondary_nodes")));
+		}
+	}
+}
+
+
 /*
  * ShardPlacementList returns the list of placements for the given shard from
  * the cache.
@@ -354,7 +564,7 @@ ShardPlacementList(uint64 shardId)
 {
 	ShardCacheEntry *shardEntry = NULL;
 	DistTableCacheEntry *tableEntry = NULL;
-	ShardPlacement *placementArray = NULL;
+	GroupShardPlacement *placementArray = NULL;
 	int numberOfPlacements = 0;
 	List *placementList = NIL;
 	int i = 0;
@@ -370,11 +580,11 @@ ShardPlacementList(uint64 shardId)
 
 	for (i = 0; i < numberOfPlacements; i++)
 	{
-		/* copy placement into target context */
-		ShardPlacement *placement = CitusMakeNode(ShardPlacement);
-		CopyShardPlacement(&placementArray[i], placement);
+		GroupShardPlacement *groupShardPlacement = &placementArray[i];
+		ShardPlacement *shardPlacement = ResolveGroupShardPlacement(groupShardPlacement,
+																	shardEntry);
 
-		placementList = lappend(placementList, placement);
+		placementList = lappend(placementList, shardPlacement);
 	}
 
 	/* if no shard placements are found, warn the user */
@@ -401,11 +611,7 @@ LookupShardCacheEntry(int64 shardId)
 
 	Assert(CitusHasBeenLoaded() && CheckCitusVersion(WARNING));
 
-	/* probably not reachable */
-	if (DistShardCacheHash == NULL)
-	{
-		InitializeDistTableCache();
-	}
+	InitializeCaches();
 
 	/* lookup cache entry */
 	shardEntry = hash_search(DistShardCacheHash, &shardId, HASH_FIND, &foundInCache);
@@ -511,10 +717,7 @@ LookupDistTableCacheEntry(Oid relationId)
 		return NULL;
 	}
 
-	if (DistTableCacheHash == NULL)
-	{
-		InitializeDistTableCache();
-	}
+	InitializeCaches();
 
 	/*
 	 * If the version is not known to be compatible, perform thorough check,
@@ -735,7 +938,7 @@ BuildCachedShardList(DistTableCacheEntry *cacheEntry)
 		cacheEntry->arrayOfPlacementArrays =
 			MemoryContextAllocZero(CacheMemoryContext,
 								   shardIntervalArrayLength *
-								   sizeof(ShardPlacement *));
+								   sizeof(GroupShardPlacement *));
 		cacheEntry->arrayOfPlacementArrayLengths =
 			MemoryContextAllocZero(CacheMemoryContext,
 								   shardIntervalArrayLength *
@@ -869,7 +1072,7 @@ BuildCachedShardList(DistTableCacheEntry *cacheEntry)
 		List *placementList = NIL;
 		MemoryContext oldContext = NULL;
 		ListCell *placementCell = NULL;
-		ShardPlacement *placementArray = NULL;
+		GroupShardPlacement *placementArray = NULL;
 		int placementOffset = 0;
 		int numberOfPlacements = 0;
 
@@ -892,35 +1095,14 @@ BuildCachedShardList(DistTableCacheEntry *cacheEntry)
 
 		/* and copy that list into the cache entry */
 		oldContext = MemoryContextSwitchTo(CacheMemoryContext);
-		placementArray = palloc0(numberOfPlacements * sizeof(ShardPlacement));
+		placementArray = palloc0(numberOfPlacements * sizeof(GroupShardPlacement));
 		foreach(placementCell, placementList)
 		{
-			ShardPlacement *srcPlacement = (ShardPlacement *) lfirst(placementCell);
-			ShardPlacement *dstPlacement = &placementArray[placementOffset];
+			GroupShardPlacement *srcPlacement =
+				(GroupShardPlacement *) lfirst(placementCell);
+			GroupShardPlacement *dstPlacement = &placementArray[placementOffset];
 
-			CopyShardPlacement(srcPlacement, dstPlacement);
-
-			/* fill in remaining fields */
-			Assert(cacheEntry->partitionMethod != 0);
-			dstPlacement->partitionMethod = cacheEntry->partitionMethod;
-			dstPlacement->colocationGroupId = cacheEntry->colocationId;
-			if (cacheEntry->partitionMethod == DISTRIBUTE_BY_HASH)
-			{
-				Assert(shardInterval->minValueExists);
-				Assert(shardInterval->valueTypeId == INT4OID);
-
-				/*
-				 * Use the lower boundary of the interval's range to identify
-				 * it for colocation purposes. That remains meaningful even if
-				 * a concurrent session splits a shard.
-				 */
-				dstPlacement->representativeValue =
-					DatumGetInt32(shardInterval->minValue);
-			}
-			else
-			{
-				dstPlacement->representativeValue = 0;
-			}
+			memcpy(dstPlacement, srcPlacement, sizeof(GroupShardPlacement));
 			placementOffset++;
 		}
 		MemoryContextSwitchTo(oldContext);
@@ -1094,7 +1276,7 @@ bool
 CitusHasBeenLoaded(void)
 {
 	/* recheck presence until citus has been loaded */
-	if (!extensionLoaded || creating_extension)
+	if (!MetadataCache.extensionLoaded || creating_extension)
 	{
 		bool extensionPresent = false;
 		bool extensionScriptExecuted = true;
@@ -1112,12 +1294,21 @@ CitusHasBeenLoaded(void)
 			{
 				extensionScriptExecuted = false;
 			}
+
+			/*
+			 * Whenever the extension exists, even when currently creating it,
+			 * we need the infrastructure to run citus in this database to be
+			 * ready.
+			 */
+			StartupCitusBackend();
 		}
 
 		/* we disable extension features during pg_upgrade */
-		extensionLoaded = extensionPresent && extensionScriptExecuted && !IsBinaryUpgrade;
+		MetadataCache.extensionLoaded = extensionPresent &&
+										extensionScriptExecuted &&
+										!IsBinaryUpgrade;
 
-		if (extensionLoaded)
+		if (MetadataCache.extensionLoaded)
 		{
 			/*
 			 * InvalidateDistRelationCacheCallback resets state such as extensionLoaded
@@ -1140,7 +1331,7 @@ CitusHasBeenLoaded(void)
 		}
 	}
 
-	return extensionLoaded;
+	return MetadataCache.extensionLoaded;
 }
 
 
@@ -1302,6 +1493,8 @@ AvailableExtensionVersion(void)
 	bool doCopy = false;
 	char *availableExtensionVersion;
 
+	InitializeCaches();
+
 	estate = CreateExecutorState();
 	extensionsResultSet = makeNode(ReturnSetInfo);
 	extensionsResultSet->econtext = GetPerTupleExprContext(estate);
@@ -1334,10 +1527,6 @@ AvailableExtensionVersion(void)
 			Datum availableVersion = slot_getattr(tupleTableSlot, 2, &isNull);
 
 			/* we will cache the result of citus version to prevent catalog access */
-			if (CacheMemoryContext == NULL)
-			{
-				CreateCacheMemoryContext();
-			}
 			oldMemoryContext = MemoryContextSwitchTo(CacheMemoryContext);
 
 			availableExtensionVersion = text_to_cstring(DatumGetTextPP(availableVersion));
@@ -1404,11 +1593,6 @@ InstalledExtensionVersion(void)
 		}
 
 		/* we will cache the result of citus version to prevent catalog access */
-		if (CacheMemoryContext == NULL)
-		{
-			CreateCacheMemoryContext();
-		}
-
 		oldMemoryContext = MemoryContextSwitchTo(CacheMemoryContext);
 
 		installedExtensionVersion = text_to_cstring(DatumGetTextPP(installedVersion));
@@ -1433,19 +1617,21 @@ InstalledExtensionVersion(void)
 Oid
 DistShardRelationId(void)
 {
-	CachedRelationLookup("pg_dist_shard", &distShardRelationId);
+	CachedRelationLookup("pg_dist_shard",
+						 &MetadataCache.distShardRelationId);
 
-	return distShardRelationId;
+	return MetadataCache.distShardRelationId;
 }
 
 
-/* return oid of pg_dist_shard_placement relation */
+/* return oid of pg_dist_placement relation */
 Oid
-DistShardPlacementRelationId(void)
+DistPlacementRelationId(void)
 {
-	CachedRelationLookup("pg_dist_shard_placement", &distShardPlacementRelationId);
+	CachedRelationLookup("pg_dist_placement",
+						 &MetadataCache.distPlacementRelationId);
 
-	return distShardPlacementRelationId;
+	return MetadataCache.distPlacementRelationId;
 }
 
 
@@ -1453,9 +1639,10 @@ DistShardPlacementRelationId(void)
 Oid
 DistNodeRelationId(void)
 {
-	CachedRelationLookup("pg_dist_node", &distNodeRelationId);
+	CachedRelationLookup("pg_dist_node",
+						 &MetadataCache.distNodeRelationId);
 
-	return distNodeRelationId;
+	return MetadataCache.distNodeRelationId;
 }
 
 
@@ -1463,9 +1650,10 @@ DistNodeRelationId(void)
 Oid
 DistLocalGroupIdRelationId(void)
 {
-	CachedRelationLookup("pg_dist_local_group", &distLocalGroupRelationId);
+	CachedRelationLookup("pg_dist_local_group",
+						 &MetadataCache.distLocalGroupRelationId);
 
-	return distLocalGroupRelationId;
+	return MetadataCache.distLocalGroupRelationId;
 }
 
 
@@ -1473,9 +1661,10 @@ DistLocalGroupIdRelationId(void)
 Oid
 DistColocationRelationId(void)
 {
-	CachedRelationLookup("pg_dist_colocation", &distColocationRelationId);
+	CachedRelationLookup("pg_dist_colocation",
+						 &MetadataCache.distColocationRelationId);
 
-	return distColocationRelationId;
+	return MetadataCache.distColocationRelationId;
 }
 
 
@@ -1484,9 +1673,9 @@ Oid
 DistColocationConfigurationIndexId(void)
 {
 	CachedRelationLookup("pg_dist_colocation_configuration_index",
-						 &distColocationConfigurationIndexId);
+						 &MetadataCache.distColocationConfigurationIndexId);
 
-	return distColocationConfigurationIndexId;
+	return MetadataCache.distColocationConfigurationIndexId;
 }
 
 
@@ -1495,9 +1684,9 @@ Oid
 DistColocationColocationidIndexId(void)
 {
 	CachedRelationLookup("pg_dist_colocation_pkey",
-						 &distColocationColocationidIndexId);
+						 &MetadataCache.distColocationColocationidIndexId);
 
-	return distColocationColocationidIndexId;
+	return MetadataCache.distColocationColocationidIndexId;
 }
 
 
@@ -1505,9 +1694,10 @@ DistColocationColocationidIndexId(void)
 Oid
 DistPartitionRelationId(void)
 {
-	CachedRelationLookup("pg_dist_partition", &distPartitionRelationId);
+	CachedRelationLookup("pg_dist_partition",
+						 &MetadataCache.distPartitionRelationId);
 
-	return distPartitionRelationId;
+	return MetadataCache.distPartitionRelationId;
 }
 
 
@@ -1516,9 +1706,9 @@ Oid
 DistPartitionLogicalRelidIndexId(void)
 {
 	CachedRelationLookup("pg_dist_partition_logical_relid_index",
-						 &distPartitionLogicalRelidIndexId);
+						 &MetadataCache.distPartitionLogicalRelidIndexId);
 
-	return distPartitionLogicalRelidIndexId;
+	return MetadataCache.distPartitionLogicalRelidIndexId;
 }
 
 
@@ -1527,9 +1717,9 @@ Oid
 DistPartitionColocationidIndexId(void)
 {
 	CachedRelationLookup("pg_dist_partition_colocationid_index",
-						 &distPartitionColocationidIndexId);
+						 &MetadataCache.distPartitionColocationidIndexId);
 
-	return distPartitionColocationidIndexId;
+	return MetadataCache.distPartitionColocationidIndexId;
 }
 
 
@@ -1538,9 +1728,9 @@ Oid
 DistShardLogicalRelidIndexId(void)
 {
 	CachedRelationLookup("pg_dist_shard_logical_relid_index",
-						 &distShardLogicalRelidIndexId);
+						 &MetadataCache.distShardLogicalRelidIndexId);
 
-	return distShardLogicalRelidIndexId;
+	return MetadataCache.distShardLogicalRelidIndexId;
 }
 
 
@@ -1548,31 +1738,32 @@ DistShardLogicalRelidIndexId(void)
 Oid
 DistShardShardidIndexId(void)
 {
-	CachedRelationLookup("pg_dist_shard_shardid_index", &distShardShardidIndexId);
+	CachedRelationLookup("pg_dist_shard_shardid_index",
+						 &MetadataCache.distShardShardidIndexId);
 
-	return distShardShardidIndexId;
+	return MetadataCache.distShardShardidIndexId;
 }
 
 
-/* return oid of pg_dist_shard_placement_shardid_index */
+/* return oid of pg_dist_placement_shardid_index */
 Oid
-DistShardPlacementShardidIndexId(void)
+DistPlacementShardidIndexId(void)
 {
-	CachedRelationLookup("pg_dist_shard_placement_shardid_index",
-						 &distShardPlacementShardidIndexId);
+	CachedRelationLookup("pg_dist_placement_shardid_index",
+						 &MetadataCache.distPlacementShardidIndexId);
 
-	return distShardPlacementShardidIndexId;
+	return MetadataCache.distPlacementShardidIndexId;
 }
 
 
-/* return oid of pg_dist_shard_placement_shardid_index */
+/* return oid of pg_dist_placement_placementid_index */
 Oid
-DistShardPlacementPlacementidIndexId(void)
+DistPlacementPlacementidIndexId(void)
 {
-	CachedRelationLookup("pg_dist_shard_placement_placementid_index",
-						 &distShardPlacementPlacementidIndexId);
+	CachedRelationLookup("pg_dist_placement_placementid_index",
+						 &MetadataCache.distPlacementPlacementidIndexId);
 
-	return distShardPlacementPlacementidIndexId;
+	return MetadataCache.distPlacementPlacementidIndexId;
 }
 
 
@@ -1580,9 +1771,10 @@ DistShardPlacementPlacementidIndexId(void)
 Oid
 DistTransactionRelationId(void)
 {
-	CachedRelationLookup("pg_dist_transaction", &distTransactionRelationId);
+	CachedRelationLookup("pg_dist_transaction",
+						 &MetadataCache.distTransactionRelationId);
 
-	return distTransactionRelationId;
+	return MetadataCache.distTransactionRelationId;
 }
 
 
@@ -1591,20 +1783,20 @@ Oid
 DistTransactionGroupIndexId(void)
 {
 	CachedRelationLookup("pg_dist_transaction_group_index",
-						 &distTransactionGroupIndexId);
+						 &MetadataCache.distTransactionGroupIndexId);
 
-	return distTransactionGroupIndexId;
+	return MetadataCache.distTransactionGroupIndexId;
 }
 
 
-/* return oid of pg_dist_shard_placement_nodeid_index */
+/* return oid of pg_dist_placement_groupid_index */
 Oid
-DistShardPlacementNodeidIndexId(void)
+DistPlacementGroupidIndexId(void)
 {
-	CachedRelationLookup("pg_dist_shard_placement_nodeid_index",
-						 &distShardPlacementNodeidIndexId);
+	CachedRelationLookup("pg_dist_placement_groupid_index",
+						 &MetadataCache.distPlacementGroupidIndexId);
 
-	return distShardPlacementNodeidIndexId;
+	return MetadataCache.distPlacementGroupidIndexId;
 }
 
 
@@ -1615,14 +1807,15 @@ CitusExtraDataContainerFuncId(void)
 	List *nameList = NIL;
 	Oid paramOids[1] = { INTERNALOID };
 
-	if (extraDataContainerFuncId == InvalidOid)
+	if (MetadataCache.extraDataContainerFuncId == InvalidOid)
 	{
 		nameList = list_make2(makeString("pg_catalog"),
 							  makeString("citus_extradata_container"));
-		extraDataContainerFuncId = LookupFuncName(nameList, 1, paramOids, false);
+		MetadataCache.extraDataContainerFuncId =
+			LookupFuncName(nameList, 1, paramOids, false);
 	}
 
-	return extraDataContainerFuncId;
+	return MetadataCache.extraDataContainerFuncId;
 }
 
 
@@ -1630,17 +1823,18 @@ CitusExtraDataContainerFuncId(void)
 Oid
 CitusWorkerHashFunctionId(void)
 {
-	if (workerHashFunctionId == InvalidOid)
+	if (MetadataCache.workerHashFunctionId == InvalidOid)
 	{
 		Oid citusExtensionOid = get_extension_oid("citus", false);
 		Oid citusSchemaOid = get_extension_schema(citusExtensionOid);
 		char *citusSchemaName = get_namespace_name(citusSchemaOid);
 		const int argCount = 1;
 
-		workerHashFunctionId = FunctionOid(citusSchemaName, "worker_hash", argCount);
+		MetadataCache.workerHashFunctionId =
+			FunctionOid(citusSchemaName, "worker_hash", argCount);
 	}
 
-	return workerHashFunctionId;
+	return MetadataCache.workerHashFunctionId;
 }
 
 
@@ -1657,11 +1851,10 @@ CitusExtensionOwner(void)
 	ScanKeyData entry[1];
 	HeapTuple extensionTuple = NULL;
 	Form_pg_extension extensionForm = NULL;
-	static Oid extensionOwner = InvalidOid;
 
-	if (extensionOwner != InvalidOid)
+	if (MetadataCache.extensionOwner != InvalidOid)
 	{
-		return extensionOwner;
+		return MetadataCache.extensionOwner;
 	}
 
 	relation = heap_open(ExtensionRelationId, AccessShareLock);
@@ -1693,8 +1886,8 @@ CitusExtensionOwner(void)
 			ereport(ERROR, (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
 							errmsg("citus extension needs to be owned by superuser")));
 		}
-		extensionOwner = extensionForm->extowner;
-		Assert(OidIsValid(extensionOwner));
+		MetadataCache.extensionOwner = extensionForm->extowner;
+		Assert(OidIsValid(MetadataCache.extensionOwner));
 	}
 	else
 	{
@@ -1706,7 +1899,7 @@ CitusExtensionOwner(void)
 
 	heap_close(relation, AccessShareLock);
 
-	return extensionOwner;
+	return MetadataCache.extensionOwner;
 }
 
 
@@ -1722,7 +1915,7 @@ CitusExtensionOwnerName(void)
 }
 
 
-/* return the  username of the currently active role */
+/* return the username of the currently active role */
 char *
 CurrentUserName(void)
 {
@@ -1732,6 +1925,99 @@ CurrentUserName(void)
 }
 
 
+/*
+ * LookupNodeRoleValueId returns the Oid of the "pg_catalog.noderole" type, or InvalidOid
+ * if it does not exist.
+ */
+static Oid
+LookupNodeRoleTypeOid()
+{
+	Value *schemaName = makeString("pg_catalog");
+	Value *typeName = makeString("noderole");
+	List *qualifiedName = list_make2(schemaName, typeName);
+	TypeName *enumTypeName = makeTypeNameFromNameList(qualifiedName);
+
+	Oid nodeRoleTypId;
+
+	/* typenameTypeId but instead of raising an error return InvalidOid */
+	Type tup = LookupTypeName(NULL, enumTypeName, NULL, false);
+	if (tup == NULL)
+	{
+		return InvalidOid;
+	}
+
+	nodeRoleTypId = HeapTupleGetOid(tup);
+	ReleaseSysCache(tup);
+
+	return nodeRoleTypId;
+}
+
+
+/*
+ * LookupNodeRoleValueId returns the Oid of the value in "pg_catalog.noderole" which
+ * matches the provided name, or InvalidOid if the noderole enum doesn't exist yet.
+ */
+static Oid
+LookupNodeRoleValueId(char *valueName)
+{
+	Oid nodeRoleTypId = LookupNodeRoleTypeOid();
+
+	if (nodeRoleTypId == InvalidOid)
+	{
+		return InvalidOid;
+	}
+	else
+	{
+		Datum nodeRoleIdDatum = ObjectIdGetDatum(nodeRoleTypId);
+		Datum valueDatum = CStringGetDatum(valueName);
+
+		Datum valueIdDatum = DirectFunctionCall2(enum_in, valueDatum, nodeRoleIdDatum);
+
+		Oid valueId = DatumGetObjectId(valueIdDatum);
+		return valueId;
+	}
+}
+
+
+/* return the Oid of the 'primary' nodeRole enum value */
+Oid
+PrimaryNodeRoleId(void)
+{
+	if (!MetadataCache.primaryNodeRoleId)
+	{
+		MetadataCache.primaryNodeRoleId = LookupNodeRoleValueId("primary");
+	}
+
+	return MetadataCache.primaryNodeRoleId;
+}
+
+
+/* return the Oid of the 'secodary' nodeRole enum value */
+Oid
+SecondaryNodeRoleId(void)
+{
+	if (!MetadataCache.secondaryNodeRoleId)
+	{
+		MetadataCache.secondaryNodeRoleId = LookupNodeRoleValueId("secondary");
+	}
+
+	return MetadataCache.secondaryNodeRoleId;
+}
+
+
+/* return the Oid of the 'unavailable' nodeRole enum value */
+Oid
+UnavailableNodeRoleId(void)
+{
+	if (!MetadataCache.unavailableNodeRoleId)
+	{
+		MetadataCache.unavailableNodeRoleId = LookupNodeRoleValueId("unavailable");
+	}
+
+	return MetadataCache.unavailableNodeRoleId;
+}
+
+
 /*
  * master_dist_partition_cache_invalidate is a trigger function that performs
  * relcache invalidations when the contents of pg_dist_partition are changed
@@ -1859,8 +2145,8 @@ master_dist_shard_cache_invalidate(PG_FUNCTION_ARGS)
 
 
 /*
- * master_dist_placmeent_cache_invalidate is a trigger function that performs
- * relcache invalidations when the contents of pg_dist_shard_placement are
+ * master_dist_placement_cache_invalidate is a trigger function that performs
+ * relcache invalidations when the contents of pg_dist_placement are
  * changed on the SQL level.
  *
  * NB: We decided there is little point in checking permissions here, there
@@ -1889,16 +2175,16 @@ master_dist_placement_cache_invalidate(PG_FUNCTION_ARGS)
 	/* collect shardid for OLD and NEW tuple */
 	if (oldTuple != NULL)
 	{
-		Form_pg_dist_shard_placement distPlacement =
-			(Form_pg_dist_shard_placement) GETSTRUCT(oldTuple);
+		Form_pg_dist_placement distPlacement =
+			(Form_pg_dist_placement) GETSTRUCT(oldTuple);
 
 		oldShardId = distPlacement->shardid;
 	}
 
 	if (newTuple != NULL)
 	{
-		Form_pg_dist_shard_placement distPlacement =
-			(Form_pg_dist_shard_placement) GETSTRUCT(newTuple);
+		Form_pg_dist_placement distPlacement =
+			(Form_pg_dist_placement) GETSTRUCT(newTuple);
 
 		newShardId = distPlacement->shardid;
 	}
@@ -1972,18 +2258,39 @@ master_dist_local_group_cache_invalidate(PG_FUNCTION_ARGS)
 }
 
 
+/*
+ * InitializeCaches() registers invalidation handlers for metadata_cache.c's
+ * caches.
+ */
+static void
+InitializeCaches(void)
+{
+	static bool performedInitialization = false;
+
+	if (!performedInitialization)
+	{
+		/* set first, to avoid recursion dangers */
+		performedInitialization = true;
+
+		/* make sure we've initialized CacheMemoryContext */
+		if (CacheMemoryContext == NULL)
+		{
+			CreateCacheMemoryContext();
+		}
+
+		InitializeDistTableCache();
+		RegisterWorkerNodeCacheCallbacks();
+		RegisterLocalGroupIdCacheCallbacks();
+	}
+}
+
+
 /* initialize the infrastructure for the metadata cache */
 static void
 InitializeDistTableCache(void)
 {
 	HASHCTL info;
 
-	/* make sure we've initialized CacheMemoryContext */
-	if (CacheMemoryContext == NULL)
-	{
-		CreateCacheMemoryContext();
-	}
-
 	/* build initial scan keys, copied for every relation scan */
 	memset(&DistPartitionScanKey, 0, sizeof(DistPartitionScanKey));
 
@@ -2037,6 +2344,14 @@ InitializeDistTableCache(void)
 HTAB *
 GetWorkerNodeHash(void)
 {
+	InitializeCaches(); /* ensure relevant callbacks are registered */
+
+	/*
+	 * Simulate a SELECT from pg_dist_node, ensure pg_dist_node doesn't change while our
+	 * caller is using WorkerNodeHash.
+	 */
+	LockRelationOid(DistNodeRelationId(), AccessShareLock);
+
 	/*
 	 * We might have some concurrent metadata changes. In order to get the changes,
 	 * we first need to accept the cache invalidation messages.
@@ -2062,19 +2377,15 @@ GetWorkerNodeHash(void)
 static void
 InitializeWorkerNodeCache(void)
 {
-	static bool invalidationRegistered = false;
 	HTAB *oldWorkerNodeHash = NULL;
 	List *workerNodeList = NIL;
 	ListCell *workerNodeCell = NULL;
 	HASHCTL info;
 	int hashFlags = 0;
 	long maxTableSize = (long) MaxWorkerNodesTracked;
+	bool includeNodesFromOtherClusters = false;
 
-	/* make sure we've initialized CacheMemoryContext */
-	if (CacheMemoryContext == NULL)
-	{
-		CreateCacheMemoryContext();
-	}
+	InitializeCaches();
 
 	/*
 	 * Create the hash that holds the worker nodes. The key is the combination of
@@ -2095,7 +2406,7 @@ InitializeWorkerNodeCache(void)
 								 &info, hashFlags);
 
 	/* read the list from pg_dist_node */
-	workerNodeList = ReadWorkerNodes();
+	workerNodeList = ReadWorkerNodes(includeNodesFromOtherClusters);
 
 	/* iterate over the worker node list */
 	foreach(workerNodeCell, workerNodeList)
@@ -2118,6 +2429,8 @@ InitializeWorkerNodeCache(void)
 		strlcpy(workerNode->workerRack, currentNode->workerRack, WORKER_LENGTH);
 		workerNode->hasMetadata = currentNode->hasMetadata;
 		workerNode->isActive = currentNode->isActive;
+		workerNode->nodeRole = currentNode->nodeRole;
+		strlcpy(workerNode->nodeCluster, currentNode->nodeCluster, NAMEDATALEN);
 
 		if (handleFound)
 		{
@@ -2132,16 +2445,20 @@ InitializeWorkerNodeCache(void)
 
 	/* now, safe to destroy the old hash */
 	hash_destroy(oldWorkerNodeHash);
+}
 
-	/* prevent multiple invalidation registrations */
-	if (!invalidationRegistered)
-	{
-		/* Watch for invalidation events. */
-		CacheRegisterRelcacheCallback(InvalidateNodeRelationCacheCallback,
-									  (Datum) 0);
 
-		invalidationRegistered = true;
-	}
+/*
+ * RegisterWorkerNodeCacheCallbacks registers the callbacks required for the
+ * worker node cache.  It's separate from InitializeWorkerNodeCache so the
+ * callback can be registered early, before the metadata tables exist.
+ */
+static void
+RegisterWorkerNodeCacheCallbacks(void)
+{
+	/* Watch for invalidation events. */
+	CacheRegisterRelcacheCallback(InvalidateNodeRelationCacheCallback,
+								  (Datum) 0);
 }
 
 
@@ -2162,6 +2479,8 @@ GetLocalGroupId(void)
 	Relation pgDistLocalGroupId = NULL;
 	Oid localGroupTableOid = InvalidOid;
 
+	InitializeCaches();
+
 	/*
 	 * Already set the group id, no need to read the heap again.
 	 */
@@ -2203,16 +2522,6 @@ GetLocalGroupId(void)
 	systable_endscan(scanDescriptor);
 	heap_close(pgDistLocalGroupId, AccessShareLock);
 
-	/* prevent multiple invalidation registrations */
-	if (!invalidationRegistered)
-	{
-		/* Watch for invalidation events. */
-		CacheRegisterRelcacheCallback(InvalidateLocalGroupIdRelationCacheCallback,
-									  (Datum) 0);
-
-		invalidationRegistered = true;
-	}
-
 	/* set the local cache variable */
 	LocalGroupId = groupId;
 
@@ -2220,6 +2529,21 @@ GetLocalGroupId(void)
 }
 
 
+/*
+ * RegisterLocalGroupIdCacheCallbacks registers the callbacks required to
+ * maintain LocalGroupId at a consistent value. It's separate from
+ * GetLocalGroupId so the callback can be registered early, before metadata
+ * tables exist.
+ */
+static void
+RegisterLocalGroupIdCacheCallbacks(void)
+{
+	/* Watch for invalidation events. */
+	CacheRegisterRelcacheCallback(InvalidateLocalGroupIdRelationCacheCallback,
+								  (Datum) 0);
+}
+
+
 /*
  * WorkerNodeHashCode computes the hash code for a worker node from the node's
  * host name and port number. Nodes that only differ by their rack locations
@@ -2276,26 +2600,12 @@ ResetDistTableCacheEntry(DistTableCacheEntry *cacheEntry)
 		 shardIndex++)
 	{
 		ShardInterval *shardInterval = cacheEntry->sortedShardIntervalArray[shardIndex];
-		ShardPlacement *placementArray = cacheEntry->arrayOfPlacementArrays[shardIndex];
-		int numberOfPlacements = cacheEntry->arrayOfPlacementArrayLengths[shardIndex];
+		GroupShardPlacement *placementArray =
+			cacheEntry->arrayOfPlacementArrays[shardIndex];
 		bool valueByVal = shardInterval->valueByVal;
 		bool foundInCache = false;
-		int placementIndex = 0;
 
 		/* delete the shard's placements */
-		for (placementIndex = 0;
-			 placementIndex < numberOfPlacements;
-			 placementIndex++)
-		{
-			ShardPlacement *placement = &placementArray[placementIndex];
-
-			if (placement->nodeName)
-			{
-				pfree(placement->nodeName);
-			}
-
-			/* placement itself is deleted as part of the array */
-		}
 		pfree(placementArray);
 
 		/* delete per-shard cache-entry */
@@ -2382,31 +2692,23 @@ InvalidateDistRelationCacheCallback(Datum argument, Oid relationId)
 	 * This happens pretty rarely, but most importantly happens during
 	 * DROP EXTENSION citus;
 	 */
-	if (relationId != InvalidOid && relationId == distPartitionRelationId)
+	if (relationId != InvalidOid && relationId == MetadataCache.distPartitionRelationId)
 	{
-		extensionLoaded = false;
-		distShardRelationId = InvalidOid;
-		distShardPlacementRelationId = InvalidOid;
-		distLocalGroupRelationId = InvalidOid;
-		distNodeRelationId = InvalidOid;
-		distColocationRelationId = InvalidOid;
-		distColocationConfigurationIndexId = InvalidOid;
-		distColocationColocationidIndexId = InvalidOid;
-		distPartitionRelationId = InvalidOid;
-		distPartitionLogicalRelidIndexId = InvalidOid;
-		distPartitionColocationidIndexId = InvalidOid;
-		distShardLogicalRelidIndexId = InvalidOid;
-		distShardShardidIndexId = InvalidOid;
-		distShardPlacementShardidIndexId = InvalidOid;
-		distShardPlacementPlacementidIndexId = InvalidOid;
-		distTransactionRelationId = InvalidOid;
-		distTransactionGroupIndexId = InvalidOid;
-		extraDataContainerFuncId = InvalidOid;
-		workerHashFunctionId = InvalidOid;
+		ClearMetadataOIDCache();
 	}
 }
 
 
+/*
+ * ClearMetadataOIDCache resets all the cached OIDs and the extensionLoaded flag.
+ */
+void
+ClearMetadataOIDCache(void)
+{
+	memset(&MetadataCache, 0, sizeof(MetadataCache));
+}
+
+
 /*
  * DistTableOidList iterates over the pg_dist_partition table and returns
  * a list that consists of the logicalrelids.
@@ -2461,7 +2763,7 @@ DistTableOidList(void)
 static void
 InvalidateNodeRelationCacheCallback(Datum argument, Oid relationId)
 {
-	if (relationId == InvalidOid || relationId == distNodeRelationId)
+	if (relationId == InvalidOid || relationId == MetadataCache.distNodeRelationId)
 	{
 		workerNodeHashValid = false;
 	}
@@ -2476,7 +2778,7 @@ static void
 InvalidateLocalGroupIdRelationCacheCallback(Datum argument, Oid relationId)
 {
 	/* when invalidation happens simply set the LocalGroupId to the default value */
-	if (relationId == InvalidOid || relationId == distLocalGroupRelationId)
+	if (relationId == InvalidOid || relationId == MetadataCache.distLocalGroupRelationId)
 	{
 		LocalGroupId = -1;
 	}
@@ -2734,13 +3036,16 @@ TupleToShardInterval(HeapTuple heapTuple, TupleDesc tupleDescriptor, Oid interva
 static void
 CachedRelationLookup(const char *relationName, Oid *cachedOid)
 {
+	/* force callbacks to be registered, so we always get notified upon changes */
+	InitializeCaches();
+
 	if (*cachedOid == InvalidOid)
 	{
 		*cachedOid = get_relname_relid(relationName, PG_CATALOG_NAMESPACE);
 
 		if (*cachedOid == InvalidOid)
 		{
-			ereport(ERROR, (errmsg("cache lookup failed for %s, called to early?",
+			ereport(ERROR, (errmsg("cache lookup failed for %s, called too early?",
 								   relationName)));
 		}
 	}
