@@ -17,6 +17,15 @@
 #include "distributed/worker_manager.h"
 #include "utils/hsearch.h"
 
+extern bool EnableVersionChecks;
+
+/* managed via guc.c */
+typedef enum
+{
+	USE_SECONDARY_NODES_NEVER = 0,
+	USE_SECONDARY_NODES_ALWAYS = 1
+} ReadFromSecondariesType;
+extern int ReadFromSecondaries;
 
 /*
  * Representation of a table's metadata that is frequently used for
@@ -36,6 +45,7 @@ typedef struct
 	bool isDistributedTable;
 	bool hasUninitializedShardInterval;
 	bool hasUniformHashDistribution; /* valid for hash partitioned tables */
+	bool hasOverlappingShardInterval;
 
 	/* pg_dist_partition metadata for this table */
 	char *partitionKeyString;
@@ -47,21 +57,43 @@ typedef struct
 	int shardIntervalArrayLength;
 	ShardInterval **sortedShardIntervalArray;
 
-	FmgrInfo *shardIntervalCompareFunction; /* NULL if no shard intervals exist */
+	/* comparator for partition column's type, NULL if DISTRIBUTE_BY_NONE */
+	FmgrInfo *shardColumnCompareFunction;
+
+	/*
+	 * Comparator for partition interval type (different from
+	 * shardValueCompareFunction if hash-partitioned), NULL if
+	 * DISTRIBUTE_BY_NONE.
+	 */
+	FmgrInfo *shardIntervalCompareFunction;
 	FmgrInfo *hashFunction; /* NULL if table is not distributed by hash */
+
+	/* pg_dist_placement metadata */
+	GroupShardPlacement **arrayOfPlacementArrays;
+	int *arrayOfPlacementArrayLengths;
 } DistTableCacheEntry;
 
 
 extern bool IsDistributedTable(Oid relationId);
 extern List * DistributedTableList(void);
 extern ShardInterval * LoadShardInterval(uint64 shardId);
+extern ShardPlacement * FindShardPlacementOnGroup(uint32 groupId, uint64 shardId);
+extern GroupShardPlacement * LoadGroupShardPlacement(uint64 shardId, uint64 placementId);
+extern ShardPlacement * LoadShardPlacement(uint64 shardId, uint64 placementId);
 extern DistTableCacheEntry * DistributedTableCacheEntry(Oid distributedRelationId);
 extern int GetLocalGroupId(void);
+extern List * DistTableOidList(void);
+extern List * ShardPlacementList(uint64 shardId);
 extern void CitusInvalidateRelcacheByRelid(Oid relationId);
 extern void CitusInvalidateRelcacheByShardId(int64 shardId);
-extern void CitusInvalidateNodeCache(void);
+extern void ClearMetadataOIDCache(void);
 
 extern bool CitusHasBeenLoaded(void);
+extern bool CheckCitusVersion(int elevel);
+extern bool CheckAvailableVersion(int elevel);
+bool MajorVersionsCompatible(char *leftVersion, char *rightVersion);
+
+extern void EnsureModificationsCanRun(void);
 
 /* access WorkerNodeHash */
 extern HTAB * GetWorkerNodeHash(void);
@@ -69,9 +101,10 @@ extern HTAB * GetWorkerNodeHash(void);
 /* relation oids */
 extern Oid DistColocationRelationId(void);
 extern Oid DistColocationConfigurationIndexId(void);
+extern Oid DistColocationColocationidIndexId(void);
 extern Oid DistPartitionRelationId(void);
 extern Oid DistShardRelationId(void);
-extern Oid DistShardPlacementRelationId(void);
+extern Oid DistPlacementRelationId(void);
 extern Oid DistNodeRelationId(void);
 extern Oid DistLocalGroupIdRelationId(void);
 
@@ -80,14 +113,20 @@ extern Oid DistPartitionLogicalRelidIndexId(void);
 extern Oid DistPartitionColocationidIndexId(void);
 extern Oid DistShardLogicalRelidIndexId(void);
 extern Oid DistShardShardidIndexId(void);
-extern Oid DistShardPlacementShardidIndexId(void);
-extern Oid DistShardPlacementPlacementidIndexId(void);
+extern Oid DistPlacementShardidIndexId(void);
+extern Oid DistPlacementPlacementidIndexId(void);
 extern Oid DistTransactionRelationId(void);
 extern Oid DistTransactionGroupIndexId(void);
-extern Oid DistShardPlacementNodeidIndexId(void);
+extern Oid DistPlacementGroupidIndexId(void);
 
 /* function oids */
 extern Oid CitusExtraDataContainerFuncId(void);
+extern Oid CitusWorkerHashFunctionId(void);
+
+/* nodeRole enum oids */
+extern Oid PrimaryNodeRoleId(void);
+extern Oid SecondaryNodeRoleId(void);
+extern Oid UnavailableNodeRoleId(void);
 
 /* user related functions */
 extern Oid CitusExtensionOwner(void);
