@@ -16,6 +16,7 @@
 
 #include "distributed/citus_nodes.h"
 #include "distributed/multi_join_order.h"
+#include "distributed/relation_restriction_equivalence.h"
 #include "nodes/nodes.h"
 #include "nodes/primnodes.h"
 #include "nodes/parsenodes.h"
@@ -24,7 +25,7 @@
 
 #define SUBQUERY_RANGE_TABLE_ID -1
 #define SUBQUERY_RELATION_ID 10000
-#define HEAP_ANALYTICS_SUBQUERY_RELATION_ID 10001
+#define SUBQUERY_PUSHDOWN_RELATION_ID 10001
 
 
 /*
@@ -180,10 +181,11 @@ extern bool SubqueryPushdown;
 
 
 /* Function declarations for building logical plans */
-extern MultiTreeRoot * MultiLogicalPlanCreate(Query *queryTree);
+extern MultiTreeRoot * MultiLogicalPlanCreate(Query *originalQuery, Query *queryTree,
+											  PlannerRestrictionContext *
+											  plannerRestrictionContext,
+											  ParamListInfo boundParams);
 extern bool NeedsDistributedPlanning(Query *queryTree);
-extern int GetRTEIdentity(RangeTblEntry *rte);
-extern void IdentifyRTE(RangeTblEntry *rte, int identifier);
 extern MultiNode * ParentNode(MultiNode *multiNode);
 extern MultiNode * ChildNode(MultiUnaryNode *multiNode);
 extern MultiNode * GrandChildNode(MultiUnaryNode *multiNode);
@@ -201,8 +203,11 @@ extern bool ExtractRangeTableIndexWalker(Node *node, List **rangeTableIndexList)
 extern List * WhereClauseList(FromExpr *fromExpr);
 extern List * QualifierList(FromExpr *fromExpr);
 extern List * TableEntryList(List *rangeTableList);
+extern List * UsedTableEntryList(Query *query);
 extern bool ExtractRangeTableRelationWalker(Node *node, List **rangeTableList);
 extern bool ExtractRangeTableEntryWalker(Node *node, List **rangeTableList);
+extern bool ExtractRangeTableRelationWalkerWithRTEExpand(Node *node,
+														 List **rangeTableList);
 extern List * pull_var_clause_default(Node *node);
 extern bool OperatorImplementsEquality(Oid opno);
 
